import itertools

class TaggedItem(object):
    """
    Float value are converted to integers
    """
    __slots__ = ['attribute', 'value']
    def __init__(self, attribute, value):
        self.attribute = attribute
        self.value = value
        if isinstance(value, float):
            self.value = int(value)
            
    def __repr__(self):
        return "%s:%s" % (self.attribute, self.value)

    def __hash__(self):
        return hash( (self.attribute, self.value) )

    def __eq__(self, other):
        return (self.attribute == other.attribute) and \
               (self.value == other.value)

    def __cmp__(self, other):
        c = cmp(self.attribute, other.attribute)
        if c != 0:
            return c
        return cmp( self.value, other.value)

class GenRule(object):
    __slots__ = ['left', 'right']

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __repr__(self):
        return "%s => %s" % (self.left, self.right)
    
def tag_transaction(transaction):
    tagged = itertools.izip(itertools.count(), transaction)
    res = [ TaggedItem(n,val) for n, val in tagged ]
    return res

class ItemsMemoizer(object):
    def __init__(self):
        self.items = {}

    def __call__(self, tagged_transaction):
        for tagged_item in tagged_transaction:
            self.items.setdefault(tagged_item.attribute, set([])).add(tagged_item)
        return tagged_transaction

class TransactionWriter(object):
    def __init__(self, fobj):
        self.fobj = fobj

    def __call__(self,tagged_transaction):
        line = ['%s' % (item,) for item in tagged_transaction]
        self.fobj.write('%s\n' % (' '.join(line),) )
        return tagged_transaction
import math


def build_ktree_from_leaves(items, attribute, k_arity, first_free_code = 0, rec = 1):
    """
    Build a k-ary tree having items as leaves
    """
    l = len(items)
    if l == 1 or k_arity < 2:
        raise StopIteration

    items.sort()
    new_items = []
    
    start = 0
    stop = k_arity
    while True:
        selection = items[start:stop]
        if len(selection) == 0:
            break
        new_item = TaggedItem(attribute = attribute, value = "%s_%s" % (rec, hex(first_free_code)))
        first_free_code += 1
        new_items.append(new_item)

        for item in selection:
            yield GenRule(item, new_item)
            
        start = stop
        stop += k_arity
    
    for rule in build_ktree_from_leaves(new_items, attribute, k_arity,
                                        first_free_code = first_free_code, rec = rec + 1):
        yield rule
    
def build_hierarchy_for_attribute(items, n_of_levels):
    """
    Build a hierarchy with n_of_levels levels having items as leaves
    """
    # for k-ary tree, leaves at depth h are k^{h}
    # thus n = k^{h} => k = math.pow(n, 1/h)
    l = len(items)
    
    if n_of_levels > 1:
        items_converging = math.pow(l, 1.0/(n_of_levels-1))
        items_converging = int(math.ceil(items_converging))
    else:
        items_converging = l
    
    items = list(items)
    attribute = items[0].attribute
    return build_ktree_from_leaves(items, attribute, items_converging)
